%% Ein Beispiel zum PageRank Algorithmus
% Der Wikipedia-Artikel <https://en.wikipedia.org/wiki/PageRank PageRank>
% zeigt ein Netzwerk von Internetseiten mit Links. Die Gre der Kreise ist
% in etwa proportional der Google-Bewertung der Seite. So wird Seite C nur
% von einer einzigen, aber gewichtigeren Seite verlinkt und hat
% infolgedessen einen hheren PageRank als Seite E, obwohl diese von
% insgesamt sechs Seiten verlinkt wird.
% 
% Das Matlab-Skript vollzieht die Berechnungen nach.
% CB 2020, 21
%% Definition der Matrix
% Ein Eintrag in Zeile i, Spalte j bedeutet : Es gibt ein Link von Seite i
% nach Seite j. Diese 01-Matrix A heisst die Adjazenzmatrix des Netzwekes.

A=[
%    A     B     C     D     E     F     G     H     I     J     K  
     0     0     0     1     0     0     0     0     0     0     0
     1     0     1     1     1     1     1     1     1     0     0
     1     1     0     0     0     0     0     0     0     0     0
     1     0     0     0     1     0     0     0     0     0     0
     1     0     0     0     0     1     1     1     1     1     1
     1     0     0     0     1     0     0     0     0     0     0
     1     0     0     0     0     0     0     0     0     0     0
     1     0     0     0     0     0     0     0     0     0     0
     1     0     0     0     0     0     0     0     0     0     0
     1     0     0     0     0     0     0     0     0     0     0
     1     0     0     0     0     0     0     0     0     0     0
];
%% 
% Die erste Spalte von A ist ein Sonderfall: Knoten A zeigt nirgendwo hin.
% Irgendwo hin klickt man aber immer, deswegen wird das so behandelt, als
% wrde A zu allen anderen Knoten hin linken.
% 
% Die Matrix wird noch geeignet skaliert.  Die Spaltensummen sollen
% einheitlich 1 sein. Dadurch lassen sich die Werte in einer Spalte j
% interpretieren als Wahrscheinlichkeit, von Seite j aus zu einer Seite i
% zu surfen.

s = sum(A);

A = A./s;  
% diese Operation funktioniert nur in neueren Matlab-Versionen.
% Alter Befehl: 
%A = bsxfun(@rdivide, A, s) % in aelteren Matlab Versionen

%% Berechnung des PageRank - mit MATLABs eig-Befehl
% Damit auch Seiten, die gar nicht oder kaum verlinkt sind, ein bisschen
% bewertet werden, bekommen alle Eintrge der Matrix noch einheitlich einen
% kleinen Wert addiert. Das wird durch den Dmpfungsfaktor d gesteuert.

d=0.85 ;
%% 
% Im Begleittext zur Abbildung in Wikipedia ist ein Wert von 85% angegeben. 
% Die Matrix mit Dmpfungsfaktor lautet dann

A_gedaempft = d*A + (1 - d)/length(A) ;
%% 
% Der Rest ist einfach: berechne einen Eigenvektor zum betragsgrten
% Eigenwert dieser Matrix und skaliere auf 100%

[U D] = eig(A_gedaempft);
diag(D) % das sind die Eigenwerte
%% 
% Einige dieser Eigenwerte sind komplexe Zahlen. Der mit grtem Betrag ist
% aber immer reell, wenn die Matrix selber nur positive Eintrge hat. (Satz
% von Frobenius).
% 
% Uns interessiert also nur der Eigenvektor zum betragsgrten Eigenwert;
% der steht in der ersten Spalte von U.

PgRnk = U(:,1)/sum(U(:,1))*100
%% 
% Das stimmt mit den Zahlenwerten der Abbildung berein.
%% PageRank iterativ berechnet
% (Dieser Algorithmus ist ein paar Milliarden Dollar wert!) Die echte
% Internet-Link-Adjazenzmatrix ist riesig gro. MATLABs eig-Befehl schafft
% das nicht. Was sich aber leistungsfhige Computer einigermassen schnell
% berechnen knnen, sind Multiplikationen Matrix*Vektor. Das ntzt der
% Page-Rank-Algorithmus aus. Im Wesentlichen ist das die Vektoriteration,
% wie sie R. von Mises und H. Pollaczek-Geiringer schon 1929 beschrieben
% haben. Allerdings noch ein bisschen einfacher: der Skalierungsschritt
% fllt weg; bei Matrizen mit Spalten-Betragssumme 1 ist der nicht
% notwendig.
% 
% 

pr = ones(length(A),1); % ein Startvektor - kann nahezu beliebig sein
pr = pr/norm(pr,1)*100  % so skaliert geben die Komponenten in Summe 100

for i=1:50
    pr = A_gedaempft*pr;
end
pr
%% 
%